from __future__ import annotations

from logging import getLogger
from pathlib import Path
from typing import TYPE_CHECKING, Callable

import boto3
import networkx as nx
from botocore.config import Config
from botocore.session import Session

from cycl.models.export_data import ExportData
from cycl.utils.cdk import get_exports_from_assembly

if TYPE_CHECKING:
    from collections.abc import Hashable


log = getLogger(__name__)


def get_graph_data(
    cdk_out_path: Path | None = None,
    aws_session: Session | None = None,
    aws_profile_name: str | None = None,
) -> dict[str, ExportData]:
    cdk_out_imports: dict[str, list[ExportData]] = (
        get_exports_from_assembly(Path(cdk_out_path)) if cdk_out_path is not None else {}
    )
    log.info('cdk_out_imports: %s', cdk_out_imports)

    # this logic should move to a module
    # profile and session should not be able to be provided
    boto_config = Config(retries={'max_attempts': 10, 'mode': 'adaptive'})
    if aws_session:
        cfn_client = aws_session.client('cloudformation', config=boto_config)  # type: ignore[attr-defined]
    elif aws_profile_name:
        cfn_client = Session(profile_name=aws_profile_name).client('cloudformation', config=boto_config)  # type: ignore[attr-defined,call-arg]
    else:
        cfn_client = boto3.client('cloudformation', config=boto_config)

    log.info('getting all exports')
    exports = ExportData.get_all_exports(cfn_client=cfn_client)
    for export_name, importing_stacks in cdk_out_imports.items():
        if export_name not in exports:
            log.warning(
                'found an export (%s) which has not been deployed yet about to be imported stack(s): (%s)',
                export_name,
                importing_stacks,
            )

    log.info('getting imports for %s exports', len(exports))
    for export in exports.values():
        if export.export_name:  # TODO: i think this is a given, maybe enforce at object level, add unit test
            export.get_all_imports(cfn_client=cfn_client)
            export.importing_stacks += cdk_out_imports.get(export.export_name, [])  # TODO: should we convert to method?
        if len(export.importing_stacks) == 0:
            log.warning('Export found with no import: %s from %s', export.export_name, export.stack_name)
    return exports


def build_graph(  # noqa: PLR0913
    graph_data: dict[str, ExportData] | None = None,
    cdk_out_path: Path | None = None,
    node_key_fn: Callable[[ExportData], Hashable] = lambda x: x.stack_name,
    nodes_to_ignore: list[str] | None = None,
    edges_to_ignore: list[list[str]] | None = None,
    aws_session: Session | None = None,
    aws_profile_name: str | None = None,
    *,
    remove_selfloops: bool = False,
) -> nx.MultiDiGraph:
    nodes_to_ignore = nodes_to_ignore or []
    edges_to_ignore = edges_to_ignore or []
    graph_data = (
        get_graph_data(cdk_out_path=cdk_out_path, aws_session=aws_session, aws_profile_name=aws_profile_name)
        if graph_data is None
        else graph_data
    )

    log.info('building dependency graph from graph data')
    dep_graph: nx.MultiDiGraph = nx.MultiDiGraph()
    for export in graph_data.values():
        export_key = node_key_fn(export)
        if export_key in nodes_to_ignore:
            continue

        edges = []
        for importing_stack in export.importing_stacks:
            importing_key = node_key_fn(importing_stack)
            if importing_key not in nodes_to_ignore:
                edge = (export_key, importing_key)
                if list(edge) not in edges_to_ignore:
                    edges.append(edge)

        if edges:
            dep_graph.add_edges_from(ebunch_to_add=edges)
        else:
            # export key
            dep_graph.add_node(export_key)

    if remove_selfloops:
        dep_graph.remove_edges_from(nx.selfloop_edges(dep_graph))

    return dep_graph



def is_circular_reversible_permutation(lst: list, candidate: list) -> bool:
    """The output cycle is non-deterministic, it could be reversed and any node could be the start."""
    extended_lst = lst + lst
    extended_lst_reverse = lst[::-1] + lst[::-1]
    return any(extended_lst[i : i + len(candidate)] == candidate for i in range(len(lst))) or any(
        extended_lst_reverse[i : i + len(candidate)] == candidate for i in range(len(lst))
    )



from importlib.metadata import PackageNotFoundError, version

from .cycl import build_graph, get_graph_data

try:
    __version__ = version('cycl')
except PackageNotFoundError:
    __version__ = 'v?'



from __future__ import annotations

import json
from logging import getLogger
from pathlib import Path

from cycl.models.export_data import ExportData

log = getLogger(__name__)


class InvalidCdkOutPathError(Exception):
    def __init__(self, message: str = 'An error occurred') -> None:
        super().__init__(message)


def __find_import_values(data: dict) -> list[str]:
    """Recursively search for all 'Fn::ImportValue' keys and return their values."""
    results = []

    if isinstance(data, dict):
        for key, value in data.items():
            if key == 'Fn::ImportValue':
                results.append(value)
            else:
                results.extend(__find_import_values(value))

    elif isinstance(data, list):
        for item in data:
            results.extend(__find_import_values(item))

    return results


def __get_import_values_from_template(file_path: Path) -> list[str]:
    """todo: handle yaml templates too."""
    with Path.open(file_path) as f:
        json_data = json.load(f)
    return __find_import_values(json_data)


def __get_stack_name_from_manifest(path_to_manifest: Path, template_file_name: str) -> str:
    """Grabs the stack name, if set, or parses stack name from the displayName (construct id) of the stack."""
    with Path.open(path_to_manifest) as f:
        json_data = json.load(f)

    artifact_id = template_file_name.split('.')[0]
    artifact = json_data['artifacts'].get(artifact_id)
    if not artifact:
        log.warning('No artifact found in manifest for %s', template_file_name)
        return ''

    stack_name = artifact.get('properties', {}).get('stackName')
    display_name_split = artifact.get('displayName', '').split('/')[-1]
    return stack_name or display_name_split


def __validate_cdk_out_path(cdk_out_path: Path) -> Path:
    cdk_out_path = Path(cdk_out_path).resolve()
    if not cdk_out_path.exists() or not cdk_out_path.is_dir():
        err_msg = f'Provided path does not exist or is not a directory: {cdk_out_path}'
        raise InvalidCdkOutPathError(err_msg)

    # handle if path is where cdk.out/ is or is directly to cdk.out
    if cdk_out_path.name != 'cdk.out':
        cdk_out_path = cdk_out_path / 'cdk.out'

    if not (cdk_out_path / 'cdk.out').exists():
        err_msg = f'File named cdk.out not found in {cdk_out_path}. Did you run `cdk synth`?'
        raise InvalidCdkOutPathError(err_msg)

    return cdk_out_path


def get_exports_from_assembly(cdk_out_path: Path) -> dict[str, list[ExportData]]:
    """Map an export name to a list of stacks which import it from the cloud assembly.

    function does not take into consideration exports defined in the assembly
        - if we found an export, we may not be able to resolve the name of it
        - AWS has built in circular dependency detection inside of a stack
        - a circular dependency couldn't be introduced in a single deployment
    """
    cdk_out_path = __validate_cdk_out_path(cdk_out_path)

    stack_import_mapping: dict[str, list[ExportData]] = {}
    for template_file in cdk_out_path.rglob('*.template.json'):
        log.info('Processing template: %s', template_file)
        imported_export_names = __get_import_values_from_template(template_file)

        log.info('found imported export names: %s', imported_export_names)
        if imported_export_names:
            manifest_path = template_file.parent / 'manifest.json'
            log.info('looking in manifest: %s', manifest_path)
            stack_name = __get_stack_name_from_manifest(manifest_path, template_file.name)
            if not stack_name:
                log.warning('unable to determine stack name for template: %s', template_file.name)
                continue
            log.info('stack name found: %s', stack_name)
            for export_name in imported_export_names:
                stack_import_mapping.setdefault(export_name, []).append(
                    ExportData(
                        stack_name=stack_name,
                        export_name=export_name,
                    )
                )
    return stack_import_mapping



import argparse
import json
import logging
import pathlib
import sys
from logging import getLogger

import networkx as nx

from cycl import build_graph
from cycl.utils.log_config import configure_log

log = getLogger(__name__)


# def get_node_key_function(node_key: str, missing_tag: str = 'ignore') -> Callable[[ExportData], str]:
#     if node_key.startswith("tag:"):
#         tag_name = node_key.split(":", 1)[1]

#         if missing_tag == 'error':
#             return lambda x: x.tags[tag_name]  # strict mode

#         default_value = f"Unknown-{tag_name}" if missing_tag == 'default' else x.stack_name
#         return lambda x: x.tags.get(tag_name, default_value)

#     node_key_map = {
#         'stack_name': lambda x: x.stack_name,
#         'export_name': lambda x: x.export_name or 'NoExport',
#         'parent_id': lambda x: x.parent_id or 'NoParent',
#     }

#     return node_key_map.get(node_key, lambda x: x.stack_name)


def create_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(prog='cycl', description='Check circular dependencies between imports and exports.')
    sp = parser.add_subparsers(dest='cmd', required=True)

    check_p = sp.add_parser('check', help='Check for cycles between AWS stack imports and exports.')
    check_p.add_argument('--exit-zero', action='store_true', help='Exit zero regardless of cyclic check result.')

    topo_p = sp.add_parser('topo', help='Find topological generations, if dependencies are acyclic')

    # global options
    for p in [check_p, topo_p]:
        p.add_argument(
            '--log-level',
            choices=['DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL'],
            default='INFO',
            help='Sets the logging level.',
        )
        p.add_argument(
            '--cdk-out',
            type=pathlib.Path,
            help='EXPERIMENTAL: Path to cdk.out/, where the cdk synthesizes the cloud assembly output.',
        )
        p.add_argument(
            '--ignore-nodes',
            nargs='+',
            default=[],
            type=str,
            help=(
                "List of nodes to to ignore when building the graph. Don't repeat ``--ignore-nodes`` if you "
                'have multiple nodes (ex. ``--ignore-nodes v1 v2``).'
            ),
        )
        p.add_argument(
            '--ignore-edge',
            nargs=2,
            default=[],
            action='append',
            metavar=('u', 'v'),
            help=(
                'Specify an edge to ignore by providing two nodes delimited by a space. ``--ignore-edge u v`` must be '
                'repeated for each edge provided.'
            ),
        )
        # p.add_argument(  # TODO
        #     '--node-key',
        #     type=str,
        #     choices=['stack_name', 'export_name', 'parent_id', 'tag:Environment'],
        #     default='stack_name',
        #     help='Specify which ExportData attribute to use as the graph node identifier.'
        # )
        # parser.add_argument(
        #     '--missing-tag',
        #     choices=['ignore', 'error', 'default'],
        #     default='ignore',
        #     help='How to handle missing tags: "ignore" (fallback to stack name), "error" (raise),
        # or "default" (custom value).'
        # )
    return parser


def app() -> None:
    parser = create_parser()

    if len(sys.argv) == 1:
        parser.print_help()
        sys.exit(0)

    args = parser.parse_args()
    configure_log(getattr(logging, args.log_level))

    dep_graph = build_graph(
        cdk_out_path=args.cdk_out,
        nodes_to_ignore=args.ignore_nodes,
        edges_to_ignore=args.ignore_edge,
    )

    cycles = list(nx.simple_cycles(dep_graph))
    for cycle in cycles:
        print(f'cycle found between nodes: {cycle}')

    if args.cmd == 'check':
        if cycles and not args.exit_zero:
            sys.exit(1)
    elif args.cmd == 'topo':
        if cycles:
            log.error('graph is cyclic, topological generations can only be computed on an acyclic graph')
            sys.exit(1)
        generations = [sorted(generation) for generation in nx.topological_generations(dep_graph)]
        print(json.dumps(generations, indent=2))
    sys.exit(0)


if __name__ == '__main__':
    app()



import logging


def configure_log(level: int) -> None:
    log_format = '%(asctime)s [%(levelname)s] %(filename)s:%(lineno)d: %(message)s'
    logging.basicConfig(level=level, format=log_format, datefmt='%Y-%m-%d %H:%M:%S')



from __future__ import annotations

from logging import getLogger
from typing import TYPE_CHECKING

import boto3
from botocore.exceptions import ClientError

if TYPE_CHECKING:
    from mypy_boto3_cloudformation import CloudFormationClient
    from mypy_boto3_cloudformation.type_defs import ListExportsOutputTypeDef

log = getLogger(__name__)


class ExportData:
    def __init__(
        self,
        stack_name: str,
        stack_id: str | None = None,
        export_name: str | None = None,
        export_value: str | None = None,
        # parent_id: str | None = None,
        # root_id: str | None = None,
        # tags: dict[str, str] | None = None,
        # outputs: list[str] | None = None,
        importing_stacks: list[ExportData] | None = None,
    ) -> None:
        self.stack_name = stack_name
        self.stack_id = stack_id
        self.export_name = export_name
        self.export_value = export_value
        # self.parent_id = parent_id
        # self.root_id = root_id
        # self.tags = tags
        # self.outputs = outputs or []
        self.importing_stacks = importing_stacks or []

    @classmethod
    def from_list_exports(cls, list_exports_resp: ListExportsOutputTypeDef) -> dict[str, ExportData]:
        """Convert an AWS CloudFormation list exports response into a dictionary of export name to ExportData instances.

        Args:
            list_exports_resp: A dictionary representing the list exports response,
                where each entry contains export details with keys:
                - 'ExportingStackId': The ID of the stack exporting the value.
                - 'Name': The name of the export.
                - 'Value': The exported value.

        Returns:
            A dictionary mapping export names to ExportData instances,
            where each ExportData object contains information about the corresponding export.
        """
        return {
            export['Name']: cls(
                stack_name=cls.parse_name_from_id(export['ExportingStackId']),
                stack_id=export['ExportingStackId'],
                export_name=export['Name'],
                export_value=export['Value'],
            )
            for export in list_exports_resp['Exports']
        }

    @classmethod
    def get_all_exports(cls, cfn_client: CloudFormationClient | None = None) -> dict[str, ExportData]:
        """Retrieve all AWS CloudFormation exports and return them as a dictionary of export name to ExportData instances.

        Args:
            cfn_client: A Boto3 CloudFormation client instance.
                If not provided, a new client will be created.

        Returns:
            A dictionary mapping export names to ExportData instances,
            where each ExportData object contains details about the corresponding export.

        Note:
            This function paginates through the AWS CloudFormation `list_exports` API to retrieve all exports.
        """
        cfn_client = cfn_client or boto3.client('cloudformation')

        exports: dict[str, ExportData] = {}
        resp = cfn_client.list_exports()
        log.debug(resp)
        exports.update(ExportData.from_list_exports(resp))
        while token := resp.get('NextToken'):
            resp = cfn_client.list_exports(NextToken=token)
            log.debug(resp)
            exports.update(ExportData.from_list_exports(resp))
        log.debug(exports)
        return exports

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, ExportData):
            return False
        return vars(self) == vars(other)

    def __hash__(self) -> int:
        return hash(tuple((key, tuple(value) if isinstance(value, list) else value) for key, value in vars(self).items()))

    def __repr__(self) -> str:
        return f'ExportData(stack_name={self.stack_name!r}, export_name={self.export_name!r})'

    def get_all_imports(self, cfn_client: CloudFormationClient | None = None) -> ExportData:
        """Retrieve all stacks that import the current export and update the importing_stacks attribute.

        Args:
            cfn_client: A Boto3 CloudFormation client instance. If not provided, a new client will be created.

        Returns:
            The updated instance with importing stacks added to the `importing_stacks` attribute.

        Note:
            This function paginates through the AWS CloudFormation `list_imports` API to retrieve all importing stacks.
            If the export is not imported by any stack, it logs a debug message instead of raising an error.
        """
        if self.export_name:
            cfn_client = cfn_client or boto3.client('cloudformation')

            try:
                resp = cfn_client.list_imports(ExportName=self.export_name)
                log.debug(resp)
                self.importing_stacks.extend(
                    [ExportData(stack_name=importing_stack_name) for importing_stack_name in resp['Imports']]
                )
                while token := resp.get('NextToken'):
                    resp = cfn_client.list_imports(ExportName=self.export_name, NextToken=token)
                    log.debug(resp)
                    self.importing_stacks.extend(
                        [ExportData(stack_name=importing_stack_name) for importing_stack_name in resp['Imports']]
                    )
            except ClientError as err:
                if 'is not imported by any stack' not in repr(err):
                    raise
                log.debug('')  # TODO: refine msg
            log.debug(self.importing_stacks)
        else:
            warning_msg = (
                f'export_name not defined for {type(self).__name__}, which probably means {self.stack_name} '
                'imports an export but does not export any'
            )
            log.warning(warning_msg)
        return self

    @staticmethod
    def parse_name_from_id(stack_id: str) -> str:
        """Extract the stack name from a given stack ID.

        Args:
            stack_id: The full stack ID, typically in the format
                'arn:aws:cloudformation:region:account-id:stack/stack-name/guid'.

        Returns:
            The extracted stack name, or an empty string if parsing fails.

        Note:
            Logs a warning if the stack ID format is unexpected.
        """
        try:
            return stack_id.split('/')[1]
        except IndexError:
            log.warning('Unable to parse name from stack_id: %s', stack_id)
            return ''
